%%%-------------------------------------------------------------------
%%% @author Krzysztof Trzepla
%%% @copyright (C) 2017 ACK CYFRONET AGH
%%% This software is released under the MIT license
%%% cited in 'LICENSE.txt'.
%%% @end
%%%-------------------------------------------------------------------
%%% @doc
%%% This file contains datastore links tests.
%%% @end
%%%-------------------------------------------------------------------
-module(datastore_links_test_SUITE).
-author("Krzysztof Trzepla").

-include("datastore_test_utils.hrl").

%% export for ct
-export([all/0, init_per_suite/1, init_per_testcase/2]).

%% tests
-export([
    add_link_should_save_link_in_memory/1,
    add_link_should_save_link_on_disc/1,
    get_link_should_return_target_from_memory/1,
    get_link_should_return_target_from_disc/1,
    delete_link_should_delete_link_in_memory/1,
    delete_link_should_delete_link_on_disc/1,
    tree_fold_should_return_all_targets/1,
    tree_fold_should_return_targets_from_offset/1,
    tree_fold_should_return_targets_from_link/1,
    tree_fold_should_return_targets_limited_by_size/1,
    tree_fold_should_return_targets_from_offset_and_limited_by_size/1,
    multi_tree_fold_should_return_all_targets/1,
    multi_tree_fold_should_return_targets_from_offset_and_limited_by_size/1,
    multi_tree_fold_should_return_targets_from_link/1,
    multi_tree_fold_should_return_all_using_ids/1,
    multi_tree_fold_should_return_all_using_ids_with_tree/1,
    multi_tree_fold_should_return_all_using_not_existing_ids/1,
    multi_tree_fold_should_return_all_using_id_with_offset/1,
    multi_tree_fold_should_return_all_using_id_with_neg_offset/1,
    multi_tree_fold_should_return_all_using_not_existsing_id_with_neg_offset/1,
    multi_sorted_tree_fold_should_return_all_using_not_existing_ids/1,
    multi_sorted_tree_fold_should_return_all_using_id_with_neg_offset/1,
    multi_sorted_tree_fold_should_return_all_using_not_existsing_id_with_neg_offset/1,
    multi_sorted_tree_fold_should_return_all_using_not_existsing_middle_id_with_neg_offset/1,
    multi_tree_fold_should_return_targets_with_token/1
]).

all() ->
    ?ALL([
        add_link_should_save_link_in_memory,
        add_link_should_save_link_on_disc,
        get_link_should_return_target_from_memory,
        get_link_should_return_target_from_disc,
        delete_link_should_delete_link_in_memory,
        delete_link_should_delete_link_on_disc,
        tree_fold_should_return_all_targets,
        tree_fold_should_return_targets_from_offset,
        tree_fold_should_return_targets_from_link,
        tree_fold_should_return_targets_limited_by_size,
        tree_fold_should_return_targets_from_offset_and_limited_by_size,
        multi_tree_fold_should_return_all_targets,
        multi_tree_fold_should_return_targets_from_offset_and_limited_by_size,
        multi_tree_fold_should_return_targets_from_link,
        multi_tree_fold_should_return_all_using_ids,
        multi_tree_fold_should_return_all_using_ids_with_tree,
        multi_tree_fold_should_return_all_using_not_existing_ids,
        multi_tree_fold_should_return_all_using_id_with_offset,
        multi_tree_fold_should_return_all_using_id_with_neg_offset,
        multi_tree_fold_should_return_all_using_not_existsing_id_with_neg_offset,
        multi_sorted_tree_fold_should_return_all_using_not_existing_ids,
        multi_sorted_tree_fold_should_return_all_using_id_with_neg_offset,
        multi_sorted_tree_fold_should_return_all_using_not_existsing_id_with_neg_offset,
        multi_sorted_tree_fold_should_return_all_using_not_existsing_middle_id_with_neg_offset,
        multi_tree_fold_should_return_targets_with_token
    ]).

-define(MODEL, ets_cached_model).
-define(CTX, #{
    model => ?MODEL,
    mutator_pid => self(),
    memory_driver => ?MEM_DRV(?MODEL),
    memory_driver_ctx => ?MEM_CTX(?MODEL),
    disc_driver => ?DISC_DRV(?MODEL),
    disc_driver_ctx => ?DISC_CTX,
    remote_driver => ?REMOTE_DRV
}).
-define(CTX(Key), datastore_multiplier:extend_name(?KEY, ?CTX)).

%%%===================================================================
%%% Test functions
%%%===================================================================

add_link_should_save_link_in_memory(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    {{ok, Link}, _} = ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, add, [
            ?LINK_NAME, ?LINK_TARGET
        ]]
    )),
    ?assertEqual(?LINK_TREE_ID, Link#link.tree_id),
    ?assertEqual(?LINK_NAME, Link#link.name),
    ?assertEqual(?LINK_TARGET, Link#link.target),
    ?assertMatch(<<_/binary>>, Link#link.tree_id).

add_link_should_save_link_on_disc(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    {{ok, Link}, _} = ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, add, [
            ?LINK_NAME, ?LINK_TARGET
        ]]
    )),
    ?assertEqual(?LINK_TREE_ID, Link#link.tree_id),
    ?assertEqual(?LINK_NAME, Link#link.name),
    ?assertEqual(?LINK_TARGET, Link#link.target),
    ?assertMatch(<<_/binary>>, Link#link.tree_id).

get_link_should_return_target_from_memory(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, add, [
            ?LINK_NAME, ?LINK_TARGET
        ]]
    )),
    {{ok, Link}, _} = ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, get, [
            ?LINK_NAME
        ]]
    )),
    ?assertEqual(?LINK_TREE_ID, Link#link.tree_id),
    ?assertEqual(?LINK_NAME, Link#link.name),
    ?assertEqual(?LINK_TARGET, Link#link.target),
    ?assertMatch(<<_/binary>>, Link#link.tree_id).

get_link_should_return_target_from_disc(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, add, [
            ?LINK_NAME, ?LINK_TARGET
        ]]
    )),
    {{ok, Link}, _} = ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, get, [
            ?LINK_NAME
        ]]
    )),
    ?assertEqual(?LINK_TREE_ID, Link#link.tree_id),
    ?assertEqual(?LINK_NAME, Link#link.name),
    ?assertEqual(?LINK_TARGET, Link#link.target),
    ?assertMatch(<<_/binary>>, Link#link.tree_id).

delete_link_should_delete_link_in_memory(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, add, [
            ?LINK_NAME, ?LINK_TARGET
        ]]
    )),
    ?assertMatch({ok, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, delete, [
            ?LINK_NAME
        ]]
    )).

delete_link_should_delete_link_on_disc(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, add, [
            ?LINK_NAME, ?LINK_TARGET
        ]]
    )),
    ?assertMatch({ok, _}, rpc:call(Worker,
        datastore_links_crud, apply, [?CTX(?KEY), ?KEY, ?LINK_TREE_ID, delete, [
            ?LINK_NAME
        ]]
    )).

tree_fold_should_return_all_targets(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    LinksNum = 1000,
    AllLinks = add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID, LinksNum),
    Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{}),
    ?assertEqual(get_expected_links(AllLinks), Links).

tree_fold_should_return_targets_from_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    LinksNum = 1000,
    AllLinks = add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID, LinksNum),
    lists:foreach(fun(Offset) ->
        Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{offset => Offset}),
        ?assertEqual(get_expected_links(AllLinks, Offset), Links)
    end, lists:seq(0, LinksNum)).

tree_fold_should_return_targets_from_link(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    LinksNum = 1000,
    AllLinks = add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID, LinksNum),
    Names = lists:map(fun(#link{name = Name}) -> Name end, AllLinks),
    lists:foreach(fun({Offset, Name}) ->
        Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{
            prev_tree_id => ?LINK_TREE_ID,
            prev_link_name => Name
        }),
        ?assertEqual(get_expected_links(AllLinks, Offset), Links)
    end, lists:zip(lists:seq(1, LinksNum), lists:sort(Names))).

tree_fold_should_return_targets_limited_by_size(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    LinksNum = 1000,
    AllLinks = add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID, LinksNum),
    lists:foreach(fun(Size) ->
        Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{size => Size}),
        ?assertEqual(get_expected_links(AllLinks, 0, Size), Links)
    end, lists:seq(0, LinksNum)).

tree_fold_should_return_targets_from_offset_and_limited_by_size(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    LinksNum = 100,
    AllLinks = add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID, LinksNum),
    lists:foreach(fun(Offset) ->
        lists:foreach(fun(Size) ->
            Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{
                offset => Offset,
                size => Size
            }),
            ?assertEqual(get_expected_links(AllLinks, Offset, Size), Links)
        end, lists:seq(0, LinksNum - Offset))
    end, lists:seq(0, LinksNum)).

multi_tree_fold_should_return_all_targets(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),
    Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{}),
    ?assertEqual(get_expected_links(AllLinks), Links).


multi_tree_fold_should_return_targets_from_offset_and_limited_by_size(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    Incr = 100,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = lists:foldl(fun(Offset, Acc) ->
        Acc ++ fold_links(Worker, ?CTX(?KEY), ?KEY, #{offset => Offset, size => Incr})
    end, [], lists:seq(0, TreesNum * LinksNum, Incr)),
    ?assertEqual(get_expected_links(AllLinks), Links).

multi_tree_fold_should_return_targets_with_token(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    Incr = 100,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = fold_links_token(Worker, ?CTX(?KEY), ?KEY,
            #{token => #link_token{}, offset => 0, size => Incr}),
    ?assertEqual(get_expected_links(AllLinks), Links).

multi_tree_fold_should_return_targets_from_link(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),
    LinkPoints = lists:map(fun(#link{tree_id = TreeId, name = Name}) ->
        {Name, TreeId}
    end, AllLinks),
    lists:foreach(fun({Offset, {Name, TreeId}}) ->
        Links = fold_links(Worker, ?CTX(?KEY), ?KEY, #{
            prev_tree_id => TreeId,
            prev_link_name => Name
        }),
        ?assertEqual(get_expected_links(AllLinks, Offset), Links)
    end, lists:zip(lists:seq(1, TreesNum * LinksNum), lists:sort(LinkPoints))).

multi_tree_fold_should_return_all_using_ids(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    Incr = 100,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = fold_links_id(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<>>}),
    ?assertEqual(get_expected_links(AllLinks), Links).

multi_tree_fold_should_return_all_using_ids_with_tree(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    Incr = 100,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = fold_links_id_and_tree(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<>>}),
    ?assertEqual(get_expected_links(AllLinks), Links).

multi_tree_fold_should_return_all_using_not_existing_ids(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 1000,
    Incr = 100,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = fold_links_non_existing_id(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<>>}),

    ?assertEqual(get_expected_links(AllLinks), Links).

multi_tree_fold_should_return_all_using_id_with_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    Incr = 100,
    Offset = 50,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = fold_links_id_and_tree(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<>>, offset => Offset}),
    Check = filter_expected_links(get_expected_links(AllLinks), [], Offset, Incr),
    ?assertEqual(Check, Links).

multi_tree_fold_should_return_all_using_id_with_neg_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 1000,
    Incr = 450,
    Offset = -250,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    ExpectedLinks = get_expected_links(AllLinks),
    #link{name = StartName} = lists:nth(abs(Offset) + 1, ExpectedLinks),
    Links = fold_links_id_and_neg_offset(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => StartName, offset => Offset}, [], false, ExpectedLinks),
    ?assertEqual(ExpectedLinks, Links),

    #link{name = StartName2} = lists:nth(abs(Offset) + 6, ExpectedLinks),
    Links2 = fold_links_id_and_neg_offset(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => StartName2, offset => Offset}, [], false, undefined),
    ?assert(length(ExpectedLinks) > length(Links2)).

multi_tree_fold_should_return_all_using_not_existsing_id_with_neg_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 1000,
    Incr = 100,
    Offset = -50,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links(Worker, ?CTX(?KEY), ?KEY, ?LINK_TREE_ID(N), LinksNum)
    end, lists:seq(1, TreesNum))),

    ExpectedLinks = get_expected_links(AllLinks),
    #link{name = StartName} = lists:nth(abs(Offset), ExpectedLinks),
    Links = fold_links_id_and_neg_offset(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<StartName/binary, ".....">>, offset => Offset},
        [], true, ExpectedLinks),
    ?assertEqual(ExpectedLinks, Links).

multi_sorted_tree_fold_should_return_all_using_not_existing_ids(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 500,
    Incr = 100,
    Offset = 50,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links_with_ids(Worker, ?CTX(?KEY), ?KEY, N, LinksNum)
    end, lists:seq(1, TreesNum))),

    Links = fold_links_non_existing_id(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<>>, offset => Offset}),
    Check = filter_expected_links(get_expected_links(AllLinks), [], Offset, Incr),
    ?assertEqual(Check, Links).

multi_sorted_tree_fold_should_return_all_using_id_with_neg_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 1000,
    Incr = 450,
    Offset = -250,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links_with_ids(Worker, ?CTX(?KEY), ?KEY, N, LinksNum)
    end, lists:seq(1, TreesNum))),

    ExpectedLinks = get_expected_links(AllLinks),
    #link{name = StartName} = lists:nth(abs(Offset) + 1, ExpectedLinks),
    Links = fold_links_id_and_neg_offset(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => StartName, offset => Offset}, [], false, ExpectedLinks),
    ?assertEqual(ExpectedLinks, Links).

multi_sorted_tree_fold_should_return_all_using_not_existsing_id_with_neg_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 1000,
    Incr = 100,
    Offset = -50,
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links_with_ids(Worker, ?CTX(?KEY), ?KEY, N, LinksNum)
    end, lists:seq(1, TreesNum))),

    ExpectedLinks = get_expected_links(AllLinks),
    #link{name = StartName} = lists:nth(abs(Offset), ExpectedLinks),
    Links = fold_links_id_and_neg_offset(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<StartName/binary, ".....">>, offset => Offset},
        [], true, ExpectedLinks),
    ?assertEqual(ExpectedLinks, Links).

multi_sorted_tree_fold_should_return_all_using_not_existsing_middle_id_with_neg_offset(Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),
    TreesNum = 5,
    LinksNum = 1000,
    Incr = 600,
    Offset = -450,
    Start = 128 * (2 * TreesNum + 1),
    AllLinks = lists:flatten(lists:map(fun(N) ->
        add_links_with_ids(Worker, ?CTX(?KEY), ?KEY, N, LinksNum)
    end, lists:seq(1, TreesNum))),

    ExpectedLinks0 = get_expected_links(AllLinks),
    #link{name = StartName} = lists:nth(Start, ExpectedLinks0),
    ExpectedLinks = lists:sublist(ExpectedLinks0, Start + Offset + 1,
        length(ExpectedLinks0) - Start - Offset),
    Links = fold_links_id_and_neg_offset(Worker, ?CTX(?KEY), ?KEY, #{size => Incr,
        prev_link_name => <<StartName/binary, ".....">>, offset => Offset},
        [], true, ExpectedLinks),
    ?assertEqual(ExpectedLinks, Links).

%%%===================================================================
%%% Init/teardown functions
%%%===================================================================

init_per_suite(Config) ->
    datastore_test_utils:init_suite([?MODEL], Config).

init_per_testcase(Case, Config) ->
    [Worker | _] = ?config(cluster_worker_nodes, Config),

    application:load(cluster_worker),
    application:set_env(cluster_worker, tp_subtrees_number, 1),
    test_utils:set_env(Worker, cluster_worker, tp_subtrees_number, 1),

    rpc:call(Worker, datastore_cache_manager, reset, [disc1]),
    rpc:call(Worker, datastore_cache_manager, resize, [disc1, cache_size(Case)]),
    Config.

%%%===================================================================
%%% Internal functions
%%%===================================================================

cache_size(add_link_should_save_link_on_disc) -> 0;
cache_size(get_link_should_return_target_from_disc) -> 0;
cache_size(delete_link_should_delete_link_on_disc) -> 0;
cache_size(_Case) -> 10000.

add_links(Worker, Ctx, Key, TreeId, LinksNum) ->
    lists:map(fun(N) ->
        {{ok, Link}, _} = ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
            datastore_links_crud, apply, [Ctx, Key, TreeId, add, [
                ?LINK_NAME(N), ?LINK_TARGET(N)
            ]]
        )),
        Link
    end, lists:seq(1, LinksNum)).

add_links_with_ids(Worker, Ctx, Key, TreeId, LinksNum) ->
    lists:map(fun(N) ->
        Name = <<"link-", (?CASE)/binary, "-",
            (integer_to_binary(10-TreeId))/binary, "-",
            (integer_to_binary(N))/binary>>,
        {{ok, Link}, _} = ?assertMatch({{ok, #link{}}, _}, rpc:call(Worker,
            datastore_links_crud, apply, [Ctx, Key, ?LINK_TREE_ID(TreeId), add, [
                Name, ?LINK_TARGET(N)
            ]]
        )),
        Link
    end, lists:seq(1, LinksNum)).

fold_links(Worker, Ctx, Key, Opts) ->
    {ok, Links} = ?assertMatch({ok, _}, rpc:call(Worker, datastore_links_iter,
        fold, [Ctx, Key, all, fun(Link, Acc) ->
            {ok, [Link | Acc]}
        end, [], Opts]
    )),
    lists:reverse(Links).

fold_links_id(Worker, Ctx, Key, #{prev_link_name := StartLink} = Opts) ->
    {ok, Links} = ?assertMatch({ok, _}, rpc:call(Worker, datastore_links_iter,
        fold, [Ctx, Key, all, fun(Link, Acc) ->
            {ok, [Link | Acc]}
        end, [], Opts]
    )),

    Filtered = lists:filter(fun(#link{name = Name}) ->
        Name =/= StartLink
    end, Links),
    case Filtered of
        [Last | _] ->
            Opts2 = Opts#{prev_link_name => Last#link.name},
            lists:reverse(Filtered) ++ fold_links_id(Worker, Ctx, Key, Opts2);
        _ ->
            []
    end.

fold_links_non_existing_id(Worker, Ctx, Key, Opts) ->
    {ok, Links} = ?assertMatch({ok, _}, rpc:call(Worker, datastore_links_iter,
        fold, [Ctx, Key, all, fun(Link, Acc) ->
            {ok, [Link | Acc]}
        end, [], Opts]
    )),

    case Links of
        [Last | _] ->
            LastBin = Last#link.name,
            Opts2 = Opts#{prev_link_name => <<LastBin/binary, ".....">>},
            lists:reverse(Links) ++ fold_links_non_existing_id(Worker, Ctx, Key, Opts2);
        _ ->
            []
    end.

fold_links_id_and_neg_offset(Worker, Ctx, Key, Opts, TmpAns, ExtendName, ExpectedLinks) ->
    {ok, Links} = ?assertMatch({ok, _}, rpc:call(Worker, datastore_links_iter,
        fold, [Ctx, Key, all, fun(Link, Acc) ->
            {ok, [Link | Acc]}
        end, [], Opts]
    )),

    Links2 = Links -- TmpAns,
    Links2Reversed = lists:reverse(Links2),

    ExpectedLinks2 = case ExpectedLinks of
        undefined ->
            ExpectedLinks;
        _ ->
            L2Len = length(Links2),
            ?assertEqual(lists:sublist(ExpectedLinks, 1, L2Len), Links2Reversed),
            lists:sublist(ExpectedLinks, L2Len + 1, length(ExpectedLinks) - L2Len)

    end,

    case Links2 of
        [Last | _] ->
            Opts2 = case ExtendName of
                true ->
                    LastBin = Last#link.name,
                    Opts#{prev_link_name => <<LastBin/binary, ".....">>};
                _ ->
                    Opts#{prev_link_name => Last#link.name}
            end,
            fold_links_id_and_neg_offset(Worker, Ctx, Key, Opts2,
                TmpAns ++ Links2Reversed, ExtendName, ExpectedLinks2);
        _ ->
            TmpAns
    end.

fold_links_id_and_tree(Worker, Ctx, Key, Opts) ->
    {ok, Links} = ?assertMatch({ok, _}, rpc:call(Worker, datastore_links_iter,
        fold, [Ctx, Key, all, fun(Link, Acc) ->
            {ok, [Link | Acc]}
        end, [], Opts]
    )),

    case Links of
        [Last | _] ->
            Reversed = lists:reverse(Links),
            Opts2 = Opts#{prev_link_name => Last#link.name,
                prev_tree_id => Last#link.tree_id},
            Reversed ++ fold_links_id_and_tree(Worker, Ctx, Key, Opts2);
        _ ->
            []
    end.

fold_links_token(Worker, Ctx, Key, Opts) ->
    {{ok, Links}, Token} = ?assertMatch({{ok, _}, _}, rpc:call(Worker, datastore_links_iter,
        fold, [Ctx, Key, all, fun(Link, Acc) ->
            {ok, [Link | Acc]}
        end, [], Opts]
    )),
    Reversed = lists:reverse(Links),
    case Token#link_token.is_last of
        true ->
            Reversed;
        _ ->
            Opts2 = Opts#{token => Token},
            Reversed ++ fold_links_token(Worker, Ctx, Key, Opts2)
    end.

get_expected_links(Links) ->
    get_expected_links(Links, 0).

get_expected_links(Links, Offset) ->
    get_expected_links(Links, Offset, length(Links)).

get_expected_links(Links, Offset, Size) ->
    SortedLinks = lists:sort(fun(
        #link{tree_id = TreeId1, name = Name1},
        #link{tree_id = TreeId2, name = Name2}
    ) ->
        {Name1, TreeId1} < {Name2, TreeId2}
    end, Links),
    lists:sublist(SortedLinks, Offset + 1, Size).

filter_expected_links(Links, Acc, Offset, Size) ->
    Length = length(Links),
    case Length =< Offset of
        true ->
            Acc;
        _ ->
            Links2 = lists:sublist(Links, Offset + 1, Length),
            case length(Links2) =< Size of
                true ->
                    Acc ++ Links2;
                _ ->
                    Links3 = lists:sublist(Links2, Size),
                    Links4 = lists:sublist(Links2, Size + 1, Length),
                    filter_expected_links(Links4, Acc ++ Links3, Offset, Size)
            end
    end.